"""
Common utility functions for NPP/SNPP
"""

import os 
import numpy as np
from pathlib import Path

# Country enumerations
EN = "en"
WA = "wa"
SC = "sc"
NI = "ni"

# Aggregations
EW = [EN, WA]
GB = [EN, WA, SC]
UK = [EN, WA, SC, NI]

# ONS country codes
CODES = {
  EN: "E92000001",
  WA: "W92000004",
  SC: "S92000003",
  NI: "N92000002"
}

def country(codes):
  """ 
  Returns country from ONS LAD code or 2-letter country code (above) - simply looks at first letter, case insensitive
  """
  if isinstance(codes, str):
    codes=[codes]
  lookup={"E": EN, "W": WA, "S": SC, "N": NI}
  #return lookup[code[0].upper()]
  raw = set([code[0] in lookup and lookup[code[0]] for code in codes])
  raw.discard(False)
  return sorted(list(raw))
  #set([code[0] in lookup and lookup[code[0]].upper() for code in codes]).discard(False))

def default_cache_dir():
  """
  Default cache dir location, ensures the path exists (failing if it cannot create) 
  This *should* work on all platforms
  """
  cache_dir = str(Path.home() / ".ukpopulation/cache")
  
  if not os.path.exists(cache_dir):
    os.makedirs(cache_dir)
  return cache_dir

def check_and_invert(categories):
  """
  Takes a list of categories to aggregrate and removes them from all possible categories, 
  creating a list of categories to preserve that can be used by groupby
  Will thow ValueError if one of supplied categories is not one of the column names below
  """
  inverted = ['C_AGE', 'GENDER', 'GEOGRAPHY_CODE', 'PROJECTED_YEAR_NAME']
  # first ensure list
  if isinstance(categories, str):
    categories = [categories]

  if "PROJECTED_YEAR_NAME" in categories:
    raise ValueError("It makes no sense to aggregate data over PROJECTED_YEAR_NAME")

  for cat in categories:
    inverted.remove(cat)

  return inverted

def filter_by_age(data, age_range):
  return data[data.C_AGE.isin(age_range)]

def aggregate(detail, categories):
  """
  Aggregate OBS_VALUE over categories
  """
  return detail.groupby(check_and_invert(categories))["OBS_VALUE"].sum().reset_index()

def split_range(full_range, cutoff):
  """
  Split a range of values into those within (<=) cutoff and those without (>)
  Returns a tuple containing 2 lists (which can be empty)
  """
  if np.isscalar(full_range):
    full_range = [full_range]

  return ([x for x in full_range if x <= cutoff], [x for x in full_range if x > cutoff])

def trim_range(input_range, minval, maxval):
  """
  Removes values < minval or > maxval from input_range
  If input_range is None, defaults to the inclusive range: range(minval, maxval + 1) 
  """
  if input_range is None:
    return range(minval, maxval + 1)
  
  if np.isscalar(input_range):
    input_range = [input_range]

  return [x for x in input_range if x >= minval and x <= maxval]

def read_cell_range(worksheet, topleft, bottomright):
  data_rows = []
  for row in worksheet[topleft:bottomright]:
    data_cols = []
    for cell in row:
      data_cols.append(cell.value)
    data_rows.append(data_cols)
  return np.array(data_rows)

def integerise(series):
  """
  This duplicates functionality that exists in humanleague, rather than intorducing a package dependency solely for this function
  """
  sumf = sum(series)
  sumi = round(sumf)
  # rescale series to nearest-integer sum
  series = series * sumi / sumf
  # get integer and fractional parts
  seriesi = np.floor(series)
  seriesf = series - seriesi
  # shortfall is integer sum less sum of integer values
  shortfall = int(sumi - sum(seriesi))
  # select the n largest fractional parts where n=shortfall 
  idx = np.argpartition(seriesf, -shortfall)[-shortfall:]
  #increment the values at these indices
  inc = np.zeros(len(series))
  np.put(inc, idx, 1)
  return seriesi + inc



import matplotlib.pyplot as plt
import ukpopulation.nppdata as NPPData
import ukpopulation.snppdata as SNPPData
import ukpopulation.utils as utils

# initialise the population modules
npp = NPPData.NPPData()
snpp = SNPPData.SNPPData()

lad = "E08000021" # Newcastle

start_year = 2016
end_year = snpp.max_year(lad)


# get the total projected population for ppp up to the SNPP horizon (2039)
ppp = snpp.aggregate(["GENDER", "C_AGE"], lad, range(start_year, end_year + 1))

# calculate the the variants
hhh = snpp.create_variant("hhh", npp, lad, range(start_year, end_year + 1))
lll = snpp.create_variant("lll", npp, lad, range(start_year, end_year + 1))

# aggregate the calculated variants by age and gender
hhh = utils.aggregate(hhh, ["GENDER", "C_AGE"])
lll = utils.aggregate(lll, ["GENDER", "C_AGE"])

# plot the data
plt.plot(ppp.PROJECTED_YEAR_NAME, ppp.OBS_VALUE, "bo", label="ppp") 
plt.plot(hhh.PROJECTED_YEAR_NAME, hhh.OBS_VALUE, "ro", label="hhh")
plt.plot(lll.PROJECTED_YEAR_NAME, lll.OBS_VALUE, "go", label="lll")
plt.xlabel("Year")
plt.ylabel("Persons")
plt.title(lad + " Population Projection Variants")
plt.legend()
plt.show()



"""
snhpdata.py - household projections
"""
import os.path
import zipfile
import pandas as pd
import requests
import tempfile
from openpyxl import load_workbook
import ukpopulation.utils as utils
import pyexcel

class SNHPData:
  """
  Functionality for downloading and collating UK Subnational Household Projection (SNHP) data
  """  

  def __init__(self, cache_dir=utils.default_cache_dir()):
    self.cache_dir = cache_dir

    self.data = {}
    self.data[utils.EN] = self.__do_england()
    self.data[utils.WA] = self.__do_wales()
    self.data[utils.SC] = self.__do_scotland()
    self.data[utils.NI] = self.__do_nireland()

  def unified(self):
    """ 
    Creates a unified dataset containing values for
    - the year range present in all datasets (2016-2039 at time of typing)
    - a lowest-common denimator set of household types (that maps to OA-level census categories)
    """
    raise NotImplementedError("The categories used in each country have no clear common denominator")
    # TODO best I can do is probably differentiate between single-person, multi-person including children, and multi-person no children households

  def min_year(self, code):
    """
    Returns the first year in the projection, assumes a single LAD or country code
    """
    # convert to country if necessary
    if "0" in code:
      code = utils.country(code)[0]
    return min(self.data[code].PROJECTED_YEAR_NAME.unique())

  def max_year(self, code):
    """
    Returns the final year in the projection, assumes a single LAD or country code
    """
    # convert to country if necessary
    if "0" in code:
      code = utils.country(code)[0]
    return max(self.data[code].PROJECTED_YEAR_NAME.unique())

  def filter(self, categories, geog_codes, years=None):
    # see unified...
    raise NotImplementedError("The categories used in each country have no clear common denominator")

  def aggregate(self, geog_codes, years=None):
    """ Returns aggregate counts of household for specified geographies and years """

    # convert geog_codes and years to arrays if single values supplied (for isin)
    if isinstance(geog_codes, str):
      geog_codes = [geog_codes]

    countries = utils.country(geog_codes)

    # TODO fix incorrect assumption is that all countries have the same year range 
    years = utils.trim_range(years, self.min_year(countries[0]), self.max_year(countries[0]))

    retval = pd.DataFrame()
    # loop over datasets as needed
    for country in countries:
      # apply filters
      retval = retval.append(self.data[country][(self.data[country].GEOGRAPHY_CODE.isin(geog_codes)) & 
                                               (self.data[country].PROJECTED_YEAR_NAME.isin(years))] \
                            ,ignore_index=True, sort=False)
    return retval.groupby(["GEOGRAPHY_CODE", "PROJECTED_YEAR_NAME"]).sum().reset_index()

  def __do_england(self):
    print("Collating SNHP data for England...")
    england_src = "https://www.ons.gov.uk/file?uri=/peoplepopulationandcommunity/populationandmigration/populationprojections/datasets/householdprojectionsforenglanddetaileddataformodellingandanalysis/2016based/detailedtablesstage1and2.zip"
    england_raw = os.path.join(self.cache_dir, os.path.basename(england_src))
    england_processed = self.cache_dir + "/snhp_e.csv"

    if os.path.isfile(england_processed): 
      snhp_e = pd.read_csv(england_processed)
    else: 
      response = requests.get(england_src)
      with open(england_raw, 'wb') as fd:
        for chunk in response.iter_content(chunk_size=1024):
          fd.write(chunk)
        print("Downloaded", england_raw)

      # this doesnt work if you directly supply the file in the zip to load_workbook
      # workaround is to extract the file to a tmp dir and load from there
      z = zipfile.ZipFile(england_raw)
      tmpdir = tempfile.TemporaryDirectory().name
      #print(tmpdir)
      z.extract("s2 Households.xlsx", tmpdir)
      sheet = load_workbook(os.path.join(tmpdir, "s2 Households.xlsx"), read_only=True)["Households"]

      raw = utils.read_cell_range(sheet,"A7", "AS32263")
      snhp_e = pd.DataFrame(raw[1:,:], columns=raw[0,:])

      # remove years before 2011 census and switch years from columns to rows
      snhp_e = snhp_e.drop([str(y) for y in range(2001,2011)], axis=1) \
        .melt(id_vars=["CODE", "AREA", "AGE GROUP", "HOUSEHOLD TYPE"]).drop("AREA", axis=1)
      # ensure count is numeric
      snhp_e.value = snhp_e.value.astype(float)
      # remove age categories and standardise column names
      snhp_e = snhp_e.groupby(["CODE", "HOUSEHOLD TYPE", "variable"]).sum().reset_index() \
        .rename({"CODE": "GEOGRAPHY_CODE", 
                "HOUSEHOLD TYPE": "HOUSEHOLD_TYPE", 
                "variable": "PROJECTED_YEAR_NAME", 
                "value": "OBS_VALUE"}, axis=1)

      snhp_e.to_csv(england_processed, index=False)

    return snhp_e

  def __do_wales(self):
    print("Collating SNHP data for Wales...")

    wales_raw = self.cache_dir + "/snhp_w.csv"
    if os.path.isfile(wales_raw): 
      snhp_w = pd.read_csv(wales_raw)
    else:
      fields = ['Area_AltCode1','Year_Code','Data','Area_Hierarchy','Variant_Code','Householdtype_ItemName_ENG']
      # StatsWales is an OData endpoint, so select fields of interest
      url = "http://open.statswales.gov.wales/dataset/hous0115?$select={}".format(",".join(fields))
      # use OData syntax to filter P (persons), AllAges (all ages), Area_Hierarchy 596 (LADs)
      url += "&$filter=Variant_Code eq 1 and Area_Hierarchy eq 'W92000004'" #Householdtype_ItemName_ENG
      data = []
      while True:
        print(url)
        r = requests.get(url)
        r.raise_for_status()
        r_data = r.json()
        data += r_data['value']
        if "odata.nextLink" in r_data:
          url = r_data["odata.nextLink"]
        else:
          break
      snhp_w = pd.DataFrame(data)

      # # Remove unwanted and rename wanted columns
      snhp_w = snhp_w.drop(["Area_Hierarchy", "Variant_Code"], axis=1)
      snhp_w = snhp_w.rename(columns={"Area_AltCode1": "GEOGRAPHY_CODE",
                                      "Data": "OBS_VALUE", 
                                      "Householdtype_ItemName_ENG": "HOUSEHOLD_TYPE",
                                      "Year_Code": "PROJECTED_YEAR_NAME"})
      # reinstate + signs that went missing
      snhp_w.HOUSEHOLD_TYPE.replace(['4 person (2 adults, 1 children)', '5 person (No children)', '5 person (2 adults, 1 children)', '5 person (1 adult, 4 children)'],
                                    ['4 person (2+ adults, 1+ children)', '5+ person (No children)', '5+ person (2+ adults, 1+ children)', '5+ person (1 adult, 4+ children)'], 
                                    inplace=True)
      # Drop occupancy and population (for now, might be useful?)
      snhp_w = snhp_w[~snhp_w.HOUSEHOLD_TYPE.isin(["Households", "Projected Private Household Population", "Average Household Size"])]
      snhp_w.to_csv(wales_raw, index=False)

    return snhp_w

  def __do_scotland(self):
    print("Collating SNHP data for Scotland...")

    scotland_processed = os.path.join(self.cache_dir, "snhp_s.csv")
    scotland_src = "https://www.nrscotland.gov.uk/files//statistics/household-projections/16/2016-house-proj-detailed-coun-princ.zip"
    scotland_raw = os.path.join(self.cache_dir, os.path.basename(scotland_src)) 

    if os.path.isfile(scotland_processed): 
      snhp_s = pd.read_csv(scotland_processed)
    else:
      response = requests.get(scotland_src)
      with open(scotland_raw, 'wb') as fd:
        for chunk in response.iter_content(chunk_size=1024):
          fd.write(chunk)
        print("Downloaded", scotland_raw)

      lookup = pd.read_csv("../microsimulation/persistent_data/gb_geog_lookup.csv.gz")
      z = zipfile.ZipFile(scotland_raw)

      snhp_s = pd.DataFrame()
      for filename in z.namelist():
        council_area = filename[36:-4].strip()
        #print(">%s<" % council_area)
        if council_area == "Na h-Eileanan Siar":
          council_area = "Comhairle nan Eilean Siar"
        elif council_area == "Scotland":
          continue

        chunk = pd.read_csv(z.open(filename, "r"), encoding="latin1", skiprows=3) \
          .dropna(how="all") \
          .drop(["Unnamed: 28", "Unnamed: 29", "Unnamed: 30"], axis=1) \
          .replace(",", "", regex=True) 
        #print(chunk)
        chunk["Unnamed: 1"] = chunk["Unnamed: 1"].str.strip()
        chunk = chunk[chunk["Unnamed: 1"] != "All ages"].drop("Unnamed: 1", axis=1)
        chunk.fillna(method='ffill', inplace=True)
        chunk["Unnamed: 0"] = chunk["Unnamed: 0"].str.strip()
        # NOTE: commas have been removed from ALL cells, so "1 adult, 1+ children" is now "1 adult 1+ children"
        chunk = chunk[chunk["Unnamed: 0"].isin(["1 adult: male", "1 adult: female", "1 adult 1+ children", "2 adults", "2+ adults 1+ children", "3+ adults", "All households"])]
        chunk = pd.melt(chunk, id_vars="Unnamed: 0")
        chunk.value = chunk.value.astype(int)
        chunk = chunk.groupby(["Unnamed: 0", "variable"]).sum().reset_index().rename({"Unnamed: 0": "HOUSEHOLD_TYPE", "variable": "PROJECTED_YEAR_NAME", "value": "OBS_VALUE"}, axis=1)
        chunk.insert(0, "GEOGRAPHY_CODE", lookup[lookup.LAD_NAME == council_area].LAD.values[0])
        snhp_s = snhp_s.append(chunk, ignore_index=True)
      snhp_s.to_csv(scotland_processed, index=False)

    return snhp_s

  def __do_nireland(self):
    # 1 worksheet per LAD equivalent
    print("Collating SNHP data for Northern Ireland...")
    ni_src = "https://www.nisra.gov.uk/sites/nisra.gov.uk/files/publications/HHP16_LGD2014.xls"
    ni_processed = os.path.join(self.cache_dir, "snhp_ni.csv")
    if os.path.isfile(ni_processed): 
      snhp_ni = pd.read_csv(ni_processed)
    else:
      ni_raw = os.path.join(self.cache_dir, os.path.basename(ni_src))
      response = requests.get(ni_src)
      with open(ni_raw, 'wb') as fd:
        for chunk in response.iter_content(chunk_size=1024):
          fd.write(chunk)

      districts = ["N090000{:02d}".format(i) for i in range(1,12)]

      # convert to temp xlsx
      tmp_xlsx_file = tempfile.NamedTemporaryFile(suffix=".xlsx").name
      #print(tmp_xlsx_file)
      pyexcel.save_book_as(file_name=ni_raw, dest_file_name=tmp_xlsx_file)
      xlsx_ni = load_workbook(tmp_xlsx_file, read_only=True)

      snhp_ni = pd.DataFrame()

      for d in districts:
        raw = utils.read_cell_range(xlsx_ni[d], "A10", "AA15") # omitting Total row
        data = pd.DataFrame(raw[1:,:], columns=raw[0,:]).melt(id_vars="Household Type*") \
          .rename({"Household Type*": "HOUSEHOLD_TYPE", "variable": "PROJECTED_YEAR_NAME", "value": "OBS_VALUE"}, axis=1)

        data.insert(0, "GEOGRAPHY_CODE", d)
        snhp_ni = snhp_ni.append(data, ignore_index=True)
        
      snhp_ni.to_csv(ni_processed, index=False)

    return snhp_ni

    


import os.path
import json
import zipfile
import numpy as np
import pandas as pd
import requests
from openpyxl import load_workbook
import ukcensusapi.Nomisweb as Api
import ukpopulation.utils as utils

class SNPPData:
  """
  Functionality for downloading and collating UK Subnational Population Projection (NPP) data
  Nomisweb stores the England data (only)
  Wales/Scotland/NI are not the responsiblity of ONS and are made avilable online by the relevant statistical agency
  """  

  def __init__(self, cache_dir=utils.default_cache_dir()):
    self.cache_dir = cache_dir
    self.data_api = Api.Nomisweb(self.cache_dir) 

    self.data = {}
    self.data[utils.EN] = self.__do_england()
    self.data[utils.WA] = self.__do_wales()
    self.data[utils.SC] = self.__do_scotland()
    self.data[utils.NI] = self.__do_nireland()

    # LADs * 26 years * 91 ages * 2 genders
    #assert len(self.data) == (326+22+32+11) * 26 * 91 * 2

  def min_year(self, code):
    """
    Returns the first year in the projection, assumes a single LAD or country code
    """
    # convert to country if necessary
    if "0" in code:
      code = utils.country(code)[0]
    return min(self.data[code].PROJECTED_YEAR_NAME.unique())

  def max_year(self, code):
    """
    Returns the final year in the projection, assumes a single LAD or country code
    """
    # convert to country if necessary
    if "0" in code:
      code = utils.country(code)[0]
    return max(self.data[code].PROJECTED_YEAR_NAME.unique())

  def filter(self, geog_codes, years=None, ages=range(0,91), genders=[1,2]):

    # convert geog_codes and years to arrays if single values supplied (for isin)
    if isinstance(geog_codes, str):
      geog_codes = [geog_codes]

    countries = utils.country(geog_codes)

    # TODO fix incorrect assumption is that all countries have the same year range 
    years = utils.trim_range(years, self.min_year(countries[0]), self.max_year(countries[0]))

    retval = pd.DataFrame() #{"GEOGRAPHY_CODE": [], "PROJECTED_YEAR_NAME": [], "C_AGE": [], "GENDER":[], "OBS_VALUE": []})
    # loop over datasets as needed
    for country in countries:
      # apply filters
      retval = retval.append(self.data[country][(self.data[country].GEOGRAPHY_CODE.isin(geog_codes)) & 
                                               (self.data[country].PROJECTED_YEAR_NAME.isin(years)) &
                                               (self.data[country].C_AGE.isin(ages)) &
                                               (self.data[country].GENDER.isin(genders))] \
                            ,ignore_index=True, sort=False)

    return retval

  def aggregate(self, categories, geog_codes, years=None, ages=range(0,91), genders=[1,2]):

    data = self.filter(geog_codes, years, ages, genders)

    # invert categories (they're the ones to aggregate, not preserve)
    return data.groupby(utils.check_and_invert(categories))["OBS_VALUE"].sum().reset_index()

  # For now one LAD at a time (due to multiple countries)
  # For now allow extrapolation of years already in data
  # Filtering age and gender is not (currently) supported
  def extrapolate(self, npp, geog_code, year_range):

    (in_range, ex_range) = utils.split_range(year_range, self.max_year(geog_code))

    all_years = self.filter(geog_code, in_range)

    for year in ex_range:
      data = self.filter([geog_code], [self.max_year(geog_code)])
      scaling = npp.year_ratio("ppp", utils.country(geog_code), self.max_year(geog_code), year)
      assert(len(data == len(scaling)))
      data.OBS_VALUE = data.OBS_VALUE * scaling.OBS_VALUE
      data.PROJECTED_YEAR_NAME = year
      all_years = all_years.append(data, ignore_index=True)

    return all_years

  def extrapolagg(self, categories, npp, geog_code, year_range):
    """
    Extrapolate and then aggregate
    """
    data = self.extrapolate(npp, geog_code, year_range)

    # invert categories (they're the ones to aggregate, not preserve)
    return data.groupby(utils.check_and_invert(categories))["OBS_VALUE"].sum().reset_index()

  def create_variant(self, variant_name, npp, geog_codes, year_range):
    """
    Apply NPP variant to SNPP: SNPP(v) = SNPP(0) * sum(a,g) [ NPP(v) / NPP(0) ]
    Preserves age-gender structure of SNPP data
    """  
    result = pd.DataFrame()
    if isinstance(geog_codes, str):
      geog_codes = [geog_codes]
    
    for geog_code in geog_codes:

      # split out any years prior to the NPP data (currently SNPP is 2014 based but NPP is 2016)
      (pre_range, in_range) = utils.split_range(year_range, npp.min_year() - 1)
      # for any years prior to NPP we just use the SNPP data as-is (i.e. "ppp")
      pre_data = self.filter(geog_code, pre_range)
      if len(pre_data) > 0:
        print("WARNING: variant {} not applied for years {} that predate the NPP data".format(variant_name, pre_range))

      # return if there's nothing in the NPP range
      if not in_range:
        result.append(pre_data)
        continue

      data = self.extrapolate(npp, geog_code, in_range).sort_values(["C_AGE", "GENDER", "PROJECTED_YEAR_NAME"]).reset_index(drop=True)

      scaling = npp.variant_ratio(variant_name, utils.country(geog_code), year_range).reset_index().sort_values(["C_AGE", "GENDER", "PROJECTED_YEAR_NAME"])
      #scaling.to_csv(variant_name + ".csv", index=False)

      #print("DF: ", len(data), ":", len(scaling))
      assert(len(data) == len(scaling))
      data.OBS_VALUE = data.OBS_VALUE * scaling.OBS_VALUE
      
      # prepend any pre-NPP data
      result = result.append(pre_data.append(data))

    return result

  def custom_variant(self, filename, checks=True, integerise=False):
    """
    Given a file containing a dataframe of the format:
      GEOGRAPHY_CODE,PEOPLE,PEOPLE_<suffix>,YEAR,net_delta
      E06000001,93019.0,93019.0,2016,0.0
      E06000002,140639.0,140639.0,2016,0.0
      E06000003,136005.0,136005.0,2016,0.0
      E06000004,196487.0,196487.0,2016,0.0
      E06000005,106347.0,106347.0,2016,0.0
      E06000006,127595.0,127595.0,2016,0.0
      E06000007,209704.0,209704.0,2016,0.0
      ...
    generate a full SNPP dataset for the supplied geographies and years, disaggregated proportionately by age and gender.
    The base variant is selected from the PEOPLE_<suffix> column name, e.g. PEOPLE_ppp would use the principal projection.
    The PEOPLE_<suffix> column should match the official variant 
    """
    custom = pd.read_csv(filename)

    # TODO get base variant and call create_variant if not ppp
    base_scenario = "ppp"
    base_column = "PEOPLE_" + base_scenario

    custom["SCALING"] = custom.PEOPLE / custom[base_column]
    #geogs = custom.index.levels[0].values
    #years = custom.index.levels[1].values
    geogs = custom.GEOGRAPHY_CODE.unique()
    years = custom.PROJECTED_YEAR_NAME.unique()

    base_dataset = self.filter(geogs, years)

    # TODO check - groupby base and check sums - fix? ignore? out-of-date MYE vs SNPP numbers
    check_dataset = self.aggregate(["GENDER", "C_AGE"], geogs, years)
    check_dataset = check_dataset.merge(custom[["GEOGRAPHY_CODE", "PROJECTED_YEAR_NAME", base_column]])
    # print(check_dataset[base_column] - check_dataset.OBS_VALUE)
    # print(check_dataset.head())
    # print(custom.head())

    base_dataset = base_dataset.merge(custom[["GEOGRAPHY_CODE", "PROJECTED_YEAR_NAME", "SCALING"]], how="left")
    base_dataset.OBS_VALUE = base_dataset.OBS_VALUE * base_dataset.SCALING
    base_dataset.rename({"SCALING": "SCALING_" + base_scenario}, axis=1, inplace=True)

    if integerise:
      base_dataset.OBS_VALUE = utils.integerise(base_dataset.OBS_VALUE)

    return base_dataset

  def __do_england(self):
    # return self.__do_england_ons() # 2014
    return self.__do_england_nomisweb() # 2016

  # nomisweb data is now 2016-based
  def __do_england_nomisweb(self):
    print("Collating SNPP data for England...")

    # need to do this in 2 batches as entire table has >1000000 rows
    table_internal = "NM_2006_1" # SNPP
    query_params = {
      "gender": "1,2",
      "c_age": "101...191",
      "MEASURES": "20100",
      "date": "latest", # 2016-based
      "projected_year": "2016...2029",
      "select": "geography_code,projected_year_name,gender,c_age,obs_value",
      "geography": "1946157057...1946157382"
    }
    snpp_e = self.data_api.get_data(table_internal, query_params)

    query_params["projected_year"] = "2030...2041"
    snpp_e = snpp_e.append(self.data_api.get_data(table_internal, query_params))
    # make age actual year
    snpp_e.C_AGE = snpp_e.C_AGE - 101

    #snpp_e[(snpp_e.GEOGRAPHY_CODE=="E08000021") & (snpp_e.PROJECTED_YEAR_NAME==2039)].to_csv("snpp_ncle_2016.csv")
    #assert(len(snpp_e) == 26*2*91*326) # 326 LADs x 91 ages x 2 genders x 26 years
    return snpp_e

  def __do_england_ons(self):
    print("Collating SNPP data for England...")
    england_src = "https://www.ons.gov.uk/file?uri=/peoplepopulationandcommunity/populationandmigration/populationprojections/datasets/localauthoritiesinenglandz1/2014based/snppz1population.zip"
    england_raw = self.cache_dir + "/snpp_e.csv"
    england_zip = self.cache_dir + "/snpp_e.zip"

    if os.path.isfile(england_raw): 
      snpp_e = pd.read_csv(england_raw)
    else: 
      response = requests.get(england_src)
      with open(england_zip, 'wb') as fd:
        for chunk in response.iter_content(chunk_size=1024):
          fd.write(chunk)
        print("Downloaded", england_zip)

      z = zipfile.ZipFile(england_zip)
      #print(z.namelist())  

      snpp_e = pd.DataFrame()
      for gender in [1,2]:
        filename = "2014 SNPP Population "+("males" if gender==1 else "females")+".csv"
        chunk = pd.read_csv(z.open(filename)) \
          .drop(["AREA_NAME", "COMPONENT", "SEX"], axis=1) \
          .query('AGE_GROUP != "All ages"')
        #.AGE_GROUP.replace({"90 and over": "90"}
        chunk.AGE_GROUP = chunk.AGE_GROUP.replace({"90 and over": "90"})
        chunk = chunk.melt(id_vars=["AREA_CODE","AGE_GROUP"])
        #chunk = chunk[chunk.AGE_GROUP != "all ages"]
        # chunk = chunk.stack().reset_index() 
        chunk.columns = ["GEOGRAPHY_CODE", "C_AGE", "PROJECTED_YEAR_NAME", "OBS_VALUE"]
        chunk["GENDER"] = gender
        snpp_e = snpp_e.append(chunk)

      #assert(len(snpp_e) == 26*2*91*326) # 326 districts x 91 ages x 2 genders x 26 years
      snpp_e.to_csv(england_raw, index=False)

    #snpp_e[(snpp_e.GEOGRAPHY_CODE=="E08000021") & (snpp_e.PROJECTED_YEAR_NAME==2039)].to_csv("snpp_ncle_2014.csv")
    return snpp_e

    # Wales
  def __do_wales(self):
    print("Collating SNPP data for Wales...")

    wales_raw = self.cache_dir + "/snpp_w.csv"
    if os.path.isfile(wales_raw): 
      snpp_w = pd.read_csv(wales_raw)
    else:
      fields = ['Area_AltCode1','Year_Code','Data','Gender_Code','Age_Code','Area_Hierarchy','Variant_Code']
      # StatsWales is an OData endpoint, so select fields of interest
      url = "http://open.statswales.gov.wales/dataset/popu5099?$select={}".format(",".join(fields))
      # use OData syntax to filter P (persons), AllAges (all ages), Area_Hierarchy 596 (LADs)
      url += "&$filter=Gender_Code ne 'P' and Area_Hierarchy eq 596 and Variant_Code eq 'Principal'"
      # 
      data = []
      while True:
        print(url)
        r = requests.get(url)
        r_data = r.json()
        data += r_data['value']
        if "odata.nextLink" in r_data:
          url = r_data["odata.nextLink"]
        else:
          break
      snpp_w = pd.DataFrame(data)

      # Remove unwanted and rename wanted columns
      snpp_w = snpp_w.drop(["Area_Hierarchy", "Variant_Code"], axis=1)
      snpp_w = snpp_w.rename(columns={"Age_Code": "C_AGE", 
                                      "Area_AltCode1": "GEOGRAPHY_CODE",
                                      "Data": "OBS_VALUE", 
                                      "Gender_Code": "GENDER", 
                                      "Year_Code": "PROJECTED_YEAR_NAME"})
      # Remove all but SYOA and make numeric 
      snpp_w = snpp_w[(snpp_w.C_AGE!="AllAges") & (snpp_w.C_AGE!="00To15") & (snpp_w.C_AGE!="16To64") & (snpp_w.C_AGE!="65Plus")]
      snpp_w.loc[snpp_w.C_AGE=="90Plus", "C_AGE"] = "90"
      snpp_w.C_AGE = pd.to_numeric(snpp_w.C_AGE)

      # convert gender to census convention 1=M, 2=F
      snpp_w.GENDER = snpp_w.GENDER.map({"M": 1, "F": 2})

      #assert(len(snpp_w) == 26*2*91*22) # 22 LADs x 91 ages x 2 genders x 26 years
      snpp_w.to_csv(wales_raw, index=False)

    return snpp_w

  def __do_scotland(self):
    print("Collating SNPP data for Scotland...")

    scotland_raw = self.cache_dir + "/snpp_s.csv"

    scotland_src = "https://www.nrscotland.gov.uk/files//statistics/population-projections/sub-national-pp-16/detailed/CA%201.zip"
    scotland_zip = self.cache_dir + "/snpp_s.zip"

    if os.path.isfile(scotland_raw): 
      snpp_s = pd.read_csv(scotland_raw)
    else: 
      response = requests.get(scotland_src)
      with open(scotland_zip, 'wb') as fd:
        for chunk in response.iter_content(chunk_size=1024):
          fd.write(chunk)
        print("Downloaded", scotland_zip)

      z = zipfile.ZipFile(scotland_zip)
      #print(z.namelist())  

      snpp_s = pd.DataFrame()
      for year in range(2016,2042):
        for gender in [1,2]:
          filename = "CA 1/Population-"+str(year)+("-Male" if gender==1 else "-Female")+".csv"
          chunk = pd.read_csv(z.open(filename)
          ).drop(["Area", "All Ages"], axis=1
          ).drop(0 
          ).rename(columns={"90 and over": "90"}
          ).set_index("Code")

          chunk = chunk.stack().reset_index() 
          chunk.columns = ["GEOGRAPHY_CODE", "C_AGE", "OBS_VALUE"]
          chunk["GENDER"] = gender
          chunk["PROJECTED_YEAR_NAME"] = year
          #print(chunk.head())
          snpp_s = snpp_s.append(chunk)

      print(len(snpp_s))
      #assert(len(snpp_s) == 26*2*91*32) # 32 districts x 91 ages x 2 genders x 26 years
      snpp_s.to_csv(scotland_raw, index=False)
    return snpp_s

  def __do_nireland(self):
    # Niron 
    # (1 worksheet per LAD equivalent)
    print("Collating SNPP data for Northern Ireland...")
    ni_src = "https://www.nisra.gov.uk/sites/nisra.gov.uk/files/publications/SNPP16_LGD14_SYA_1641.xlsx"
    ni_raw = self.cache_dir + "/snpp_ni.csv"
    if os.path.isfile(ni_raw): 
      snpp_ni = pd.read_csv(ni_raw)
    else:
      response = requests.get(ni_src)
      with open(self.cache_dir + "/ni_raw.xlsx", 'wb') as fd:
        for chunk in response.iter_content(chunk_size=1024):
          fd.write(chunk)

      # easier to hard-code the worksheet names we need (since unlikely to change frequently)
      districts=["Antrim & Newtownabbey",
                "Ards & North Down",
                "Armagh Banbridge & Craigavon",
                "Belfast",
                "Causeway Coast & Glens",
                "Derry & Strabane",
                "Fermanagh & Omagh",
                "Lisburn & Castlereagh",
                "Mid & East Antrim",
                "Mid Ulster",
                "Newry Mourne & Down"]

      xls_ni = load_workbook(self.cache_dir + "/ni_raw.xlsx", read_only=True)

      snpp_ni = pd.DataFrame()

      for d in districts:
        # 1 extra row compared to 2014 data (below was A2)
        area_code = xls_ni[d]["A3"].value
        # 2 extra rows compared to 2014 data (below was A3:A95)
        males = utils.read_cell_range(xls_ni[d], "A5", "AA97")
        females = utils.read_cell_range(xls_ni[d], "A100", "AA192")
        
        dfm = pd.DataFrame(data=males[1:,1:], index=males[1:,0], columns=males[0,1:]).drop(["Age"]).stack().reset_index()
        dfm.columns=["C_AGE", "PROJECTED_YEAR_NAME", "OBS_VALUE"]
        dfm["GENDER"] = pd.Series(1, dfm.index)
        dfm["GEOGRAPHY_CODE"] = pd.Series(area_code, dfm.index)
        dfm.loc[dfm.C_AGE=="90+", "C_AGE"] = "90"

        dff = pd.DataFrame(data=females[1:,1:], index=females[1:,0], columns=females[0,1:]).drop(["Age"]).stack().reset_index()
        dff.columns=["C_AGE", "PROJECTED_YEAR_NAME", "OBS_VALUE"]
        dff["GENDER"] = pd.Series(2, dff.index)
        dff["GEOGRAPHY_CODE"] = pd.Series(area_code, dff.index)
        dff.loc[dff.C_AGE=="90+", "C_AGE"] = 90

        snpp_ni = snpp_ni.append(dfm)
        snpp_ni = snpp_ni.append(dff)

      #assert(len(snpp_ni) == 26*2*91*11) # 11 districts x 91 ages x 2 genders x 26 years
      snpp_ni.to_csv(ni_raw, index=False)

    return snpp_ni



import matplotlib.pyplot as plt
import ukpopulation.myedata as MYEData
import ukpopulation.nppdata as NPPData
import ukpopulation.snppdata as SNPPData
import ukpopulation.utils as utils

# initialise the population modules
mye = MYEData.MYEData()
npp = NPPData.NPPData()
snpp = SNPPData.SNPPData()

lad = "E07000041" # Exeter

# 50 year horizon
years = range(2011, 2067)

(mye_years, proj_years) = utils.split_range(years, mye.max_year())
(snpp_years, npp_years) = utils.split_range(proj_years, snpp.max_year(lad))

pop_mye = mye.aggregate(["GENDER", "C_AGE"], lad, mye_years)
# get the total projected population for pop up to the SNPP horizon (2039)
pop = snpp.aggregate(["GENDER", "C_AGE"], lad, snpp_years)
# extrapolate for another ~25 years
pop_ex = snpp.extrapolagg(["GENDER", "C_AGE"], npp, lad, npp_years)

# plot the data
plt.plot(pop_mye.PROJECTED_YEAR_NAME, pop_mye.OBS_VALUE, "b^", label="MYE")
plt.plot(pop.PROJECTED_YEAR_NAME, pop.OBS_VALUE, "bo", label="SNPP")
plt.plot(pop_ex.PROJECTED_YEAR_NAME, pop_ex.OBS_VALUE, "b.", label="ext NPP")
plt.xlabel("Year")
plt.ylabel("Persons")
plt.legend()
plt.title(lad + " Extrapolated Population (Principal Variant)")
plt.show()




import os.path
import requests
import zipfile
import time
import numpy as np
import pandas as pd
from openpyxl import load_workbook
import ukcensusapi.Nomisweb as Api
from bs4 import BeautifulSoup
import ukpopulation.utils as utils

def _read_excel_xml(path, sheet_name):
  file = open(path, encoding="utf8")
  soup = BeautifulSoup(file, 'xml')
  worksheet = []
  for sheet in soup.find_all('Worksheet'):
    # workaround for an intermittent issue where the "ss:" prefix disappears from the attribute keys
    # Probably related to this xml namespace issue: 
    # https://stackoverflow.com/questions/37195992/beautifulsoup4-removes-namespace-definitions-from-schema-in-wsdl
    if sheet.get("ss:Name") == sheet_name or sheet.get("Name") == sheet_name:
      for row in sheet.findAll('Row'):
        row_as_list = []
        for cell in row.findAll('Cell'):
          data = cell.find('Data')
          if data:
            row_as_list.append(data.text)
        worksheet.append(row_as_list)
  return worksheet

class NPPData:
  """
  Functionality for downloading and collating UK National Population Projection (NPP) data, including variants
  Nomisweb stores the UK principal variant (only)
  Other variants are avilable online in zipped xml files
  """

  VARIANTS = {
    "hhh": "High population", 
    "hpp": "High fertility",
    "lll": "Low population",
    "lpp": "Low fertility",
    "php": "High life expectancy",
    "pjp": "Moderately high life expectancy",
    "pkp": "Moderately low life expectancy",
    "plp": "Low life expectancy",
    "pph": "High migration",
    "ppl": "Low migration",
    "ppp": "Principal",
    "ppq": "0% future EU migration (non-ONS)",
    "ppr": "50% future EU migration (non-ONS)",
    "pps": "150% future EU migration (non-ONS)",
    "ppz": "Zero net migration"
  }
  # Other variants not in data?
  # Young age structure 	hlh 				
  # Old age structure 	lhl 				
  # Replacement fertility 	rpp 				
  # Constant fertility 	cpp 				
  # No mortality improvement 	pnp 				
  # No change 	cnp 				
  # Long term balanced net migration 	ppb 	

  def __init__(self, cache_dir = None):
    if cache_dir is None:
      cache_dir = utils.default_cache_dir()
    self.cache_dir = cache_dir
    self.data_api = Api.Nomisweb(self.cache_dir) 
    # map of pandas dataframes keyed by variant code
    self.data = {}

    # load principal aggressively...
    self.data["ppp"] = self.__download_ppp()

    # ...and variants lazily
    #self.__download_variants()

  def min_year(self):
    """
    Returns the first year in the projection
    """
    return min(self.data["ppp"].PROJECTED_YEAR_NAME.unique())

  def max_year(self):
    """
    Returns the final year in the projection
    """
    return max(self.data["ppp"].PROJECTED_YEAR_NAME.unique())


  def detail(self, variant_name, geog, years=None, ages=range(0,91), genders=[1,2]):
    """
    Return a subset of the raw data
    """
    if not variant_name in NPPData.VARIANTS:
      raise RuntimeError("invalid variant name: " + variant_name)
    # make years a valid range (this *silently* removes invalid years)
    years = utils.trim_range(years, self.min_year(), self.max_year())

    if not variant_name in self.data:
      # is it a standard variant
      if variant_name in NPPData.VARIANTS:
        self.__load_variant(variant_name)
      else:
        raise RuntimeError("Invalid variant name / custom variants are not yet implemented")

    # apply filters
    if isinstance(geog, str):
      geog = [geog]
    geog_codes = [utils.CODES[g] for g in geog]
    return self.data[variant_name][(self.data[variant_name].GEOGRAPHY_CODE.isin(geog_codes)) & 
                                   (self.data[variant_name].PROJECTED_YEAR_NAME.isin(years)) &
                                   (self.data[variant_name].C_AGE.isin(ages)) &
                                   (self.data[variant_name].GENDER.isin(genders))].reset_index(drop=True)


  def aggregate(self, categories, variant_name, geog, years=None, ages=range(0,91), genders=[1,2]):
    """
    Subset and aggregate the raw data
    """

    data = self.detail(variant_name, geog, years, ages, genders)

    return data.groupby(utils.check_and_invert(categories))["OBS_VALUE"].sum().reset_index()

  def year_ratio(self, variant_name, geog, ref_year, year, ages=range(0,91), genders=[1,2]):
    """
    Ratio to ref_year projection for selected geog/years/ages/genders 
    """
    # TODO do we need to (re)index these to ensure the piecewise mult is correct?
    ref = self.detail(variant_name, geog, [ref_year], ages, genders)
    num = self.detail(variant_name, geog, [year], ages, genders)

    assert(len(ref) == len(num))
    num.OBS_VALUE = num.OBS_VALUE / ref.OBS_VALUE
    return num

  def variant_ratio(self, variant_numerator, geog, years, ages=range(0,91), genders=[1,2]): 
    """
    Ratio to principal projection for selected geog/years/ages/genders 
    """
    # this function only works for a single country (which is ok)
    ref = self.detail("ppp", geog, years, ages, genders).set_index(["C_AGE", "GENDER", "PROJECTED_YEAR_NAME"])

    num = self.detail(variant_numerator, geog, years, ages, genders).set_index(["C_AGE", "GENDER", "PROJECTED_YEAR_NAME"])

    assert(len(ref) == len(num))
    num.OBS_VALUE = num.OBS_VALUE / ref.OBS_VALUE
    #print(num.query("GENDER==2 and C_AGE==6 and PROJECTED_YEAR_NAME==2018"))

    # return multiindexed df
    return num

  def force_load_variants(self, variants):

    for variant in variants:
      if not variant in self.data:
        self.__load_variant(variant)
    return 

  def __download_ppp(self):

    print("Loading NPP principal (ppp) data for England, Wales, Scotland & Northern Ireland")

    table_internal = "NM_2009_1" # 2016-based NPP (principal)
    query_params = {
      "gender": "1,2",
      "c_age": "1...106",
      "MEASURES": "20100",
      "date": "latest",
      "projected_year": "2016...2116",
      "select": "geography_code,projected_year_name,gender,c_age,obs_value",
      "geography": "2092957699...2092957702"
    }
    ppp = self.data_api.get_data(table_internal, query_params)
    # make age actual year
    ppp.C_AGE = ppp.C_AGE - 1
    # TODO merge 90+ for consistency with SNPP
    pop90plus = ppp[ppp.C_AGE >= 90].groupby(["GENDER", "PROJECTED_YEAR_NAME", "GEOGRAPHY_CODE"])["OBS_VALUE"].sum().reset_index()
    pop90plus["C_AGE"] = 90

    # remove the aggregated categories from the original and append the aggregate
    ppp = ppp[ppp.C_AGE < 90].append(pop90plus, ignore_index=True, sort=False)

    return ppp
  
  def __load_variant(self, variant_name):

    # [4 country zips] -> [60 xml] -> [60 raw csv] -> [15 variant csv]

    datasets = {
      utils.EN: "https://www.ons.gov.uk/file?uri=/peoplepopulationandcommunity/populationandmigration/populationprojections/datasets/z3zippedpopulationprojectionsdatafilesengland/2016based/tablez3opendata16england.zip",
      utils.WA: "https://www.ons.gov.uk/file?uri=/peoplepopulationandcommunity/populationandmigration/populationprojections/datasets/z4zippedpopulationprojectionsdatafileswales/2016based/tablez4opendata16wales.zip",
      utils.SC: "https://www.ons.gov.uk/file?uri=/peoplepopulationandcommunity/populationandmigration/populationprojections/datasets/z5zippedpopulationprojectionsdatafilesscotland/2016based/tablez5opendata16scotland.zip",
      utils.NI: "https://www.ons.gov.uk/file?uri=/peoplepopulationandcommunity/populationandmigration/populationprojections/datasets/z6zippedpopulationprojectionsdatafilesnorthernireland/2016based/tablez6opendata16northernireland.zip",
    }

    # check for cached data
    dataset = self.cache_dir + "/npp_" + variant_name + ".csv"
    if not os.path.isfile(dataset):
      # assemble data if not already cached
      self.data[variant_name] = pd.DataFrame()

      # step 1: download, country-level zip file containing all variants (if not already there)
      for country in datasets:
        raw_zip = self.cache_dir + "/npp_" + country + ".zip"
        if not os.path.isfile(raw_zip): 
          print("downloading " + raw_zip)
          response = requests.get(datasets[country])
          with open(raw_zip, 'wb') as fd:
            for chunk in response.iter_content(chunk_size=1024):
              fd.write(chunk)   
        else:
          print("using " + raw_zip)

      for country in datasets:
        raw_zip = self.cache_dir + "/npp_" + country + ".zip"
        z = zipfile.ZipFile(raw_zip)
        # step 2: unzip, collate and reformat data if not presentcd
        print("Extracting " + country + "_" + variant_name)
        vxml = country + "_" + variant_name + "_opendata2016.xml"
        if not os.path.isfile(self.cache_dir + "/" + vxml):
          z.extract(vxml, path=self.cache_dir)
        start = time.time()
        vdata = np.array(_read_excel_xml(str(self.cache_dir + "/" + vxml), "Population"))
        print("read xml in " + str(time.time() - start))
        df = pd.DataFrame(data=vdata[1:,0:], columns=vdata[0,0:])
        df = df.set_index(["Sex", "Age"]).stack().reset_index()
        # remove padding
        df.Age = df.Age.str.strip()
        df.columns = ["GENDER", "C_AGE", "PROJECTED_YEAR_NAME", "OBS_VALUE"]
        #print(df.head())

        # list age categories we are aggregating
        a = ["90", "91", "92", "93", "94", "95", "96", "97", "98", "99", "100", "101", "102", "103", "104", "105 - 109", "110 and over"]

        # copy the data in these categories
        dfagg = df[df.C_AGE.isin(a)]
        #print(dfagg.head())

        # The aggregration goes haywire unless the data is saved and reloaded - comment out lines below to see
        # TODO this needs to be fixed and/or reported as a (reproducible) bug
        tmpfile = self.cache_dir + "/tmp.csv"
        dfagg.to_csv(tmpfile, index=False)
        dfagg = pd.read_csv(tmpfile)

        dfagg = dfagg.groupby(["GENDER", "PROJECTED_YEAR_NAME"])["OBS_VALUE"].sum().reset_index()
        dfagg["C_AGE"] = "90"
        #print(dfagg.head())

        # print(df.head())
        # print(dfagg.head())
        # print(df.columns)
        # print(dfagg.columns)
        # remove the aggregated categories from the original and append the aggregate
        df = df[~df.C_AGE.isin(a)].append(dfagg, ignore_index=True)

        # add the country code
        df["GEOGRAPHY_CODE"] = utils.CODES[country]

        #df.to_csv(vcsv, index=None)
        self.data[variant_name] = self.data[variant_name].append(df, ignore_index=True)
      
      # step 3: save preprocessed data
      self.data[variant_name].to_csv(dataset, index=None)

    self.data[variant_name] = pd.read_csv(dataset)



import matplotlib.pyplot as plt
import ukpopulation.nppdata as NPPData
import ukpopulation.snppdata as SNPPData
import ukpopulation.utils as utils

# initialise the population modules
npp = NPPData.NPPData()
snpp = SNPPData.SNPPData()

years = range(snpp.min_year(), snpp.max_year())

variant = "hhh"
region = "E09000001"

npp.force_load_variants(npp.VARIANTS)
for k in npp.data:
  print("NPP ", k, ":", npp.data[k].C_AGE.unique())

print("SNPP:", snpp.data.C_AGE.unique())

for year in years:
  if year <= snpp.max_year():
  #age_sex = Utils.create_age_sex_marginal(self.snpp[self.snpp.PROJECTED_YEAR_NAME == year], self.region)
  # print(self.snpp[self.snpp.PROJECTED_YEAR_NAME == year])
  # print(self.snpp_api.create_variant(self.variant, self.npp_api, self.region, year))
    age_sex = snpp.create_variant(variant, npp, region, year)
  else:
    age_sex = snpp.extrapolate(npp, region, year)


"""
MYEData - wrapper around Mid-Year Estimate data by LAD, SYoA and gender
"""

import pandas as pd
import ukcensusapi.Nomisweb as Api
import ukpopulation.utils as utils

class MYEData:
  """
  Functionality for downloading and collating UK mid-year estimate (MYE) data
  Nomisweb stores the data for the entire UK, from 1991-2016 inclusive
  """  

  # update as and when necessary (this is unlike (S)NPP where we query the data for the year range)
  # the data is stored differently at nomisweb (year is part of the query) 
  MIN_YEAR = 1991
  MAX_YEAR = 2016

  def __init__(self, cache_dir=None):
    if cache_dir is None:
      cache_dir = utils.default_cache_dir()
    self.cache_dir = cache_dir
    self.data_api = Api.Nomisweb(self.cache_dir) 

    # store as a dictionary keyed by year (lazy retrieval)
    self.data = {}

  def min_year(self):
    """
    Returns the first year in the data
    """
    return MYEData.MIN_YEAR

  def max_year(self):
    """
    Returns the final year in the data
    """
    return MYEData.MAX_YEAR

  # TODO functionality for easy aggregration to E/W/EW/S/GB/NI/UK

  def filter(self, geog_codes, years=None, ages=range(0,91), genders=[1,2]):
    """
    Get MYE detailed data for a given year
    """
    # ensure array inputs
    if isinstance(geog_codes, str):
      geog_codes = [geog_codes]
    if isinstance(ages, int):
      ages = [ages]
    if isinstance(genders, int):
      genders = [genders]

    result = pd.DataFrame()

    years = utils.trim_range(years, self.min_year(), self.max_year())

    for year in years:

      # ensure the data is loaded
      self.__fetch_data(year)

      ## ensure we return a copy!
      part = self.data[year][(self.data[year].GEOGRAPHY_CODE.isin(geog_codes)) &
                             (self.data[year].C_AGE.isin(ages)) &
                             (self.data[year].GENDER.isin(genders))].copy()
      part["PROJECTED_YEAR_NAME"] = year
      result = result.append(part)

    return result.reset_index(drop=True)

  def aggregate(self, categories, geog_codes, years=None, ages=range(0,91), genders=[1,2]):

    data = self.filter(geog_codes, years, ages, genders)

    # invert categories (they're the ones to aggregate, not preserve)
    return data.groupby(utils.check_and_invert(categories))["OBS_VALUE"].sum().reset_index()

  def __fetch_data(self, year):
    """
    Gets Mid-year population estimate data for a given year
    Data is by single year of age by gender by local authority
    """
    # if data already loaded return 
    if year in self.data:
      return

    table_internal = "NM_2002_1" # 2016-based MYE
    query_params = {
      "gender": "1,2",
      "c_age": "101...191",
      "MEASURES": "20100",
      "select": "geography_code,gender,c_age,obs_value",
      "geography": "1879048193...1879048573,1879048583,1879048574...1879048582"
    }

    if year < MYEData.MIN_YEAR or year > MYEData.MAX_YEAR:
      raise ValueError("{} is outside the available years for MYE data ({}-{})".format(year, MIN_YEAR, MAX_YEAR))

    query_params["date"] = "latest"
    if year < MYEData.MAX_YEAR:
      query_params["date"] += "MINUS" + str(2016-year)

    self.data[year] = self.data_api.get_data(table_internal, query_params)

    # renumber age so that 0 means [0,1)
    self.data[year].C_AGE -= 101

    return self.data[year]




#!/usr/bin/env python3

import setuptools

def readme():
  with open('README.md') as f:
    return f.read()

setuptools.setup(
  name='ukpopulation',
  version='1.1.0',
  description='Download, cache, collate, filter, manipulate and extrapolate UK population and household estimates/projections',
  long_description=readme(),
  long_description_content_type="text/markdown",
  url='https://github.com/nismod/ukpopulation',
  author='Andrew P Smith, Tom Russell',
  author_email='a.p.smith@leeds.ac.uk, tom.russell@ouce.ox.ac.uk',
  packages=setuptools.find_packages(),
  install_requires=["numpy",
                    "pandas",
                    "requests",
                    "openpyxl",
                    "beautifulsoup4",
                    "lxml",
                    "pyexcel",
                    "pyexcel-xls",
                    "pyexcel-xlsx",
                    "ukcensusapi"],
  classifiers=(
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
  ),
  test_suite='nose.collector',
  tests_require=['nose']
)



import matplotlib.pyplot as plt
import ukpopulation.nppdata as NPPData
import ukpopulation.snppdata as SNPPData
import ukpopulation.utils as utils

# initialise the population modules
npp = NPPData.NPPData()
snpp = SNPPData.SNPPData()

# 50 years, roughly half is extrapolated
years = range(2016, 2067)

lad = "E08000021" # Newcastle

# get the total projected (principal variant) population for ppp, extrapolating as necessary
ppp = snpp.extrapolagg(["GENDER", "C_AGE"], npp, lad, years)

# calculate the the variants
hhh = snpp.create_variant("hhh", npp, lad, years)
lll = snpp.create_variant("lll", npp, lad, years)

# aggregate the calculated variants by age and gender
hhh = utils.aggregate(hhh, ["GENDER", "C_AGE"])
lll = utils.aggregate(lll, ["GENDER", "C_AGE"])

# plot the data
plt.plot(ppp.PROJECTED_YEAR_NAME[:24], ppp.OBS_VALUE[:24], "bo", label="ppp") 
plt.plot(ppp.PROJECTED_YEAR_NAME[24:], ppp.OBS_VALUE[24:], "b.", label="ppp (ext)") 
plt.plot(hhh.PROJECTED_YEAR_NAME[:24], hhh.OBS_VALUE[:24], "ro", label="hhh")
plt.plot(hhh.PROJECTED_YEAR_NAME[24:], hhh.OBS_VALUE[24:], "r.", label="hhh (ext)")
plt.plot(lll.PROJECTED_YEAR_NAME[:24], lll.OBS_VALUE[:24], "go", label="lll")
plt.plot(lll.PROJECTED_YEAR_NAME[24:], lll.OBS_VALUE[24:], "g.", label="lll (ext)")
plt.xlabel("Year")
plt.ylabel("Persons")
plt.title(lad + " Extrapolated Population Projection Variants")
plt.legend()
plt.show()



import pandas as pd
import ukpopulation.nppdata as NPPData
import ukpopulation.snppdata as SNPPData
import ukpopulation.utils as utils

# initialise the population modules
npp = NPPData.NPPData()
snpp = SNPPData.SNPPData()

# start with an empty data frame
result = pd.DataFrame()

country = utils.WA
horizon = 2050

# get the first year where extrapolation is necessary, extrapolate to 2050
ex_years = range(snpp.max_year(country) + 1, horizon + 1)


# loop over all the Welsh LADs 
for lad in snpp.data[country].GEOGRAPHY_CODE.unique():
  print(lad)

  # extrapolate and aggregate
  lad = snpp.extrapolagg(["GENDER", "C_AGE"], npp, lad, ex_years)
  # append to data
  result = result.append(lad, ignore_index=True)

# write out results
result.to_csv("snpp_extrap_{}_{}.csv".format(country, horizon), index=False)

